@@ -1,4 +1,4 @@
-/* dnsmasq is Copyright (c) 2000-2007 Simon Kelley
+/* dnsmasq is Copyright (c) 2000-2009 Simon Kelley
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -10,8 +10,8 @@
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
      
-  You should have received a copy of the GNU General Public License
-  along with this program.  If not, see <http://www.gnu.org/licenses/>.
+   You should have received a copy of the GNU General Public License
+   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
 
 #include "dnsmasq.h"
@@ -302,6 +302,7 @@ static int forward_query(int udpfd, union mysockaddr *udpaddr,
 		      if (!forward->rfd6 &&
 			  !(forward->rfd6 = allocate_rfd(AF_INET6)))
 			break;
+		      daemon->rfd_save = forward->rfd6;
 		      fd = forward->rfd6->fd;
 		    }
 		  else
@@ -310,6 +311,7 @@ static int forward_query(int udpfd, union mysockaddr *udpaddr,
 		      if (!forward->rfd4 &&
 			  !(forward->rfd4 = allocate_rfd(AF_INET)))
 			break;
+		      daemon->rfd_save = forward->rfd4;
 		      fd = forward->rfd4->fd;
 		    }
 		}
@@ -379,8 +381,8 @@ static size_t process_reply(HEADER *header, time_t now,
   size_t plen; 
 
   /* If upstream is advertising a larger UDP packet size
-	 than we allow, trim it so that we don't get overlarge
-	 requests for the client. We can't do this for signed packets. */
+     than we allow, trim it so that we don't get overlarge
+     requests for the client. We can't do this for signed packets. */
 
   if ((pheader = find_pseudoheader(header, n, &plen, &sizep, &is_sign)) && !is_sign)
     {
@@ -916,19 +918,23 @@ static struct randfd *allocate_rfd(int family)
      (eg) TFTP. Once we have a reasonable number, randomness should be OK */
 
   for (i = 0; i < RANDOM_SOCKS; i++)
-    if (daemon->randomsocks[i].refcount == 0 && 
-	(daemon->randomsocks[i].fd = random_sock(family)) != -1)
+    if (daemon->randomsocks[i].refcount == 0)
       {
+	if ((daemon->randomsocks[i].fd = random_sock(family)) == -1)
+	  break;
+      
 	daemon->randomsocks[i].refcount = 1;
 	daemon->randomsocks[i].family = family;
 	return &daemon->randomsocks[i];
       }
 
-  /* No free ones, grab an existing one */
+  /* No free ones or cannot get new socket, grab an existing one */
   for (i = 0; i < RANDOM_SOCKS; i++)
     {
       int j = (i+finger) % RANDOM_SOCKS;
-      if (daemon->randomsocks[j].family == family && daemon->randomsocks[j].refcount != 0xffff)
+      if (daemon->randomsocks[j].refcount != 0 &&
+	  daemon->randomsocks[j].family == family && 
+	  daemon->randomsocks[j].refcount != 0xffff)
 	{
 	  finger = j;
 	  daemon->randomsocks[j].refcount++;
